// Copyright 2016 PDFium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// Original code copyright 2014 Foxit Software Inc. http://www.foxitsoftware.com

#ifndef CORE_FPDFAPI_FPDF_PAGE_CPDF_GRAPHSTATE_H_
#define CORE_FPDFAPI_FPDF_PAGE_CPDF_GRAPHSTATE_H_

#include "core/fxge/include/fx_ge.h"

class CPDF_GraphState : public CFX_CountRef<CFX_GraphStateData> {};

#endif  // CORE_FPDFAPI_FPDF_PAGE_CPDF_GRAPHSTATE_H_
